@@ -2,7 +2,10 @@
 # (Yes, it has a point!)
 
 
-__copyright__ = "Copyright (C) 2009-2013 Andreas Kloeckner"
+__copyright__ = """
+Copyright (C) 2009-2013 Andreas Kloeckner
+Copyright (C) 2020 Matt Wala
+"""
 
 __license__ = """
 Permission is hereby granted, free of charge, to any person obtaining a copy
@@ -31,10 +34,10 @@ import operator
 import sys
 import logging
 from typing import (
-        Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, TypeVar)
+        cast, Any, Callable, Dict, Hashable, Iterable,
+        List, Optional, Set, Tuple, TypeVar)
 import builtins
 
-
 from sys import intern
 
 
@@ -92,9 +95,7 @@ Permutations, Tuples, Integer sequences
 .. autofunction:: wandering_element
 .. autofunction:: generate_nonnegative_integer_tuples_below
 .. autofunction:: generate_nonnegative_integer_tuples_summing_to_at_most
-.. autofunction:: generate_all_nonnegative_integer_tuples
 .. autofunction:: generate_all_integer_tuples_below
-.. autofunction:: generate_all_integer_tuples
 .. autofunction:: generate_permutations
 .. autofunction:: generate_unique_permutations
 
@@ -102,6 +103,7 @@ Formatting
 ----------
 
 .. autoclass:: Table
+.. autofunction:: merge_tables
 .. autofunction:: string_histogram
 .. autofunction:: word_wrap
 
@@ -167,6 +169,12 @@ Hashing
 
 .. autofunction:: unordered_hash
 
+Sampling
+--------
+
+.. autofunction:: sphere_sample_equidistant
+.. autofunction:: sphere_sample_fibonacci
+
 Type Variables Used
 -------------------
 
@@ -250,7 +258,7 @@ def deprecate_keyword(oldkey: str,
 # }}}
 
 
-# {{{ math --------------------------------------------------------------------
+# {{{ math
 
 def delta(x, y):
     if x == y:
@@ -369,7 +377,7 @@ class RecordWithoutPickling:
     def __repr__(self):
         return "{}({})".format(
                 self.__class__.__name__,
-                ", ".join("{}={!r}".format(fld, getattr(self, fld))
+                ", ".join(f"{fld}={getattr(self, fld)!r}"
                     for fld in self.__class__.fields
                     if hasattr(self, fld)))
 
@@ -466,7 +474,7 @@ class FakeList:
             return self._Function(index)
 
 
-# {{{ dependent dictionary ----------------------------------------------------
+# {{{ dependent dictionary
 
 class DependentDictionary:
     def __init__(self, f, start=None):
@@ -613,8 +621,8 @@ def memoize(*args: F, **kwargs: Any) -> F:
 
     if kwargs:
         raise TypeError(
-            "memoize received unexpected keyword arguments: %s"
-            % ", ".join(list(kwargs.keys())))
+            "memoize received unexpected keyword arguments: {}".format(
+                ", ".join(kwargs.keys())))
 
     if key_func is not None:
         def _decorator(func):
@@ -660,7 +668,7 @@ def memoize(*args: F, **kwargs: Any) -> F:
     if callable(args[0]) and len(args) == 1:
         return _decorator(args[0])
     raise TypeError(
-        "memoize received unexpected position arguments: %s" % args)
+        f"memoize received unexpected position arguments: {args}")
 
 
 FunctionValueCache = memoize
@@ -670,7 +678,7 @@ class _HasKwargs:
     pass
 
 
-def memoize_on_first_arg(function, cache_dict_name=None):
+def memoize_on_first_arg(function: F, cache_dict_name: Optional[str] = None) -> F:
     """Like :func:`memoize_method`, but for functions that take the object
     in which do memoization information is stored as first argument.
 
@@ -708,9 +716,12 @@ def memoize_on_first_arg(function, cache_dict_name=None):
 
     from functools import update_wrapper
     new_wrapper = update_wrapper(wrapper, function)
-    new_wrapper.clear_cache = clear_cache
 
-    return new_wrapper
+    # type-ignore because mypy has a point here, stuffing random attributes
+    # into the function's dict is moderately sketchy.
+    new_wrapper.clear_cache = clear_cache  # type: ignore[attr-defined]
+
+    return cast(F, new_wrapper)
 
 
 def memoize_method(method: F) -> F:
@@ -890,7 +901,7 @@ class memoize_in:  # noqa
         e.g. frozen :mod:`dataclasses`.
     """
 
-    def __init__(self, container, identifier):
+    def __init__(self, container: Any, identifier: Hashable) -> None:
         try:
             memoize_in_dict = container._pytools_memoize_in_dict
         except AttributeError:
@@ -900,7 +911,7 @@ class memoize_in:  # noqa
 
         self.cache_dict = memoize_in_dict.setdefault(identifier, {})
 
-    def __call__(self, inner):
+    def __call__(self, inner: F) -> F:
         @wraps(inner)
         def new_inner(*args):
             try:
@@ -910,7 +921,10 @@ class memoize_in:  # noqa
                 self.cache_dict[args] = result
                 return result
 
-        return new_inner
+        # NOTE: mypy gets confused because it types `wraps` as
+        #   Callable[[VarArg(Any)], Any]
+        # which, for some reason, is not compatible with `F`
+        return new_inner                # type: ignore[return-value]
 
 
 class keyed_memoize_in:  # noqa
@@ -1003,9 +1017,11 @@ def negate_tuple(t1):
 
 
 def shift(vec, dist):
-    """Return a copy of C{vec} shifted by C{dist}.
+    """Return a copy of *vec* shifted by *dist* such that
 
-    @postcondition: C{shift(a, i)[j] == a[(i+j) % len(a)]}
+    .. code:: python
+
+        shift(a, i)[j] == a[(i+j) % len(a)]
     """
 
     result = vec[:]
@@ -1100,7 +1116,7 @@ def reverse_dictionary(the_dict):
     for key, value in the_dict.items():
         if value in result:
             raise RuntimeError(
-                    "non-reversible mapping, duplicate key '%s'" % value)
+                    f"non-reversible mapping, duplicate key '{value}'")
         result[value] = key
     return result
 
@@ -1155,7 +1171,7 @@ def find_max_where(predicate, prec=1e-5, initial_guess=1, fail_bound=1e38):
 
             if mag > fail_bound:
                 raise RuntimeError("predicate appears to be true "
-                        "everywhere, up to %g" % fail_bound)
+                        f"everywhere, up to {fail_bound:g}")
 
         lower_true = mag/2
         upper_false = mag
@@ -1432,22 +1448,8 @@ def generate_nonnegative_integer_tuples_summing_to_at_most(n, length):
                 yield remainder + (i,)
 
 
-def generate_all_nonnegative_integer_tuples(length, least=0):
-    assert length >= 0
-    current_max = least
-    while True:
-        for max_pos in range(length):
-            for prebase in generate_nonnegative_integer_tuples_below(
-                    current_max, max_pos, least):
-                for postbase in generate_nonnegative_integer_tuples_below(
-                        current_max+1, length-max_pos-1, least):
-                    yield prebase + [current_max] + postbase
-        current_max += 1
-
-
 # backwards compatibility
 generate_positive_integer_tuples_below = generate_nonnegative_integer_tuples_below
-generate_all_positive_integer_tuples = generate_all_nonnegative_integer_tuples
 
 
 def _pos_and_neg_adaptor(tuple_iter):
@@ -1467,11 +1469,6 @@ def generate_all_integer_tuples_below(n, length, least_abs=0):
         n, length, least_abs))
 
 
-def generate_all_integer_tuples(length, least_abs=0):
-    return _pos_and_neg_adaptor(generate_all_nonnegative_integer_tuples(
-        length, least_abs))
-
-
 def generate_permutations(original):
     """Generate all permutations of the list *original*.
 
@@ -1582,29 +1579,66 @@ a_star = MovedFunctionDeprecationWrapper(a_star_moved)
 class Table:
     """An ASCII table generator.
 
-    :arg alignments: List of alignments of each column ('l', 'c', or 'r',
+    .. attribute:: nrows
+    .. attribute:: ncolumns
+
+    .. attribute:: alignments
+
+        A :class:`tuple` of alignments of each column: ``"l"``, ``"c"``, or ``"r"``,
         for left, center, and right alignment, respectively). Columns which
         have no alignment specifier will use the last specified alignment. For
-        example, with `alignments=['l', 'r']`, the third and all following
-        columns will use 'r' alignment.
+        example, with ``alignments=("l", "r")``, the third and all following
+        columns will use right alignment.
 
     .. automethod:: add_row
+
     .. automethod:: __str__
-    .. automethod:: latex
     .. automethod:: github_markdown
+    .. automethod:: csv
+    .. automethod:: latex
     """
 
-    def __init__(self, alignments=None):
-        self.rows = []
-        if alignments is not None:
-            self.alignments = alignments
+    def __init__(self, alignments: Optional[Tuple[str, ...]] = None) -> None:
+        if alignments is None:
+            alignments = ("l",)
         else:
-            self.alignments = ["l"]
+            if any(a not in ("l", "c", "r") for a in alignments):
+                raise ValueError(f"alignments are ('l', 'c', 'r'): {alignments}")
 
-    def add_row(self, row):
-        self.rows.append([str(i) for i in row])
+            alignments = tuple(alignments)
 
-    def __str__(self):
+        self.rows: List[Tuple[str, ...]] = []
+        self.alignments = alignments
+
+    @property
+    def nrows(self) -> int:
+        return len(self.rows)
+
+    @property
+    def ncolumns(self) -> int:
+        return len(self.rows[0])
+
+    def add_row(self, row: Tuple[Any, ...]) -> None:
+        if self.rows and len(row) != self.ncolumns:
+            raise ValueError(
+                    f"tried to add a row with {len(row)} columns to "
+                    f"a table with {self.ncolumns} columns")
+
+        self.rows.append(tuple([str(i) for i in row]))
+
+    def _get_alignments(self) -> Tuple[str, ...]:
+        # NOTE: If not all alignments were specified, extend alignments with the
+        # last alignment specified
+        return (self.alignments
+                + (self.alignments[-1],) * (self.ncolumns - len(self.alignments))
+                )
+
+    def _get_column_widths(self, rows) -> Tuple[int, ...]:
+        return tuple([
+            max(len(row[i]) for row in rows) for i in range(self.ncolumns)
+            ])
+
+    def __str__(self) -> str:
         """
         Returns a string representation of the table.
 
@@ -1619,15 +1653,11 @@ class Table:
             10 | 20||
 
         """
+        if not self.rows:
+            return ""
 
-        columns = len(self.rows[0])
-        col_widths = [max(len(row[i]) for row in self.rows)
-                      for i in range(columns)]
-
-        alignments = self.alignments
-        # If not all alignments were specified, extend alignments with the
-        # last alignment specified:
-        alignments += self.alignments[-1] * (columns - len(self.alignments))
+        alignments = self._get_alignments()
+        col_widths = self._get_column_widths(self.rows)
 
         lines = [" | ".join([
             cell.center(col_width) if align == "c"
@@ -1640,7 +1670,7 @@ class Table:
 
         return "\n".join(lines)
 
-    def github_markdown(self):
+    def github_markdown(self) -> str:
         r"""Returns a string representation of the table formatted as
         `GitHub-Flavored Markdown.
         <https://docs.github.com/en/github/writing-on-github/organizing-information-with-tables>`__
@@ -1656,17 +1686,16 @@ class Table:
             10 | 20\|\|
 
         """  # noqa: W605
-        # Pipe symbols ('|') must be replaced
-        rows = [[w.replace("|", "\\|") for w in r] for r in self.rows]
+        if not self.rows:
+            return ""
 
-        columns = len(rows[0])
-        col_widths = [max(len(row[i]) for row in rows)
-                      for i in range(columns)]
+        def escape(cell: str) -> str:
+            # Pipe symbols ('|') must be replaced
+            return cell.replace("|", "\\|")
 
-        alignments = self.alignments
-        # If not all alignments were specified, extend alignments with the
-        # last alignment specified:
-        alignments += self.alignments[-1] * (columns - len(self.alignments))
+        rows = [tuple([escape(cell) for cell in row]) for row in self.rows]
+        alignments = self._get_alignments()
+        col_widths = self._get_column_widths(rows)
 
         lines = [" | ".join([
             cell.center(col_width) if align == "c"
@@ -1675,14 +1704,16 @@ class Table:
             for cell, col_width, align in zip(row, col_widths, alignments)])
             for row in rows]
         lines[1:1] = ["|".join(
-            ":" + "-" * (col_width - 1 + (i > 0)) + ":" if align == "c"
-            else ":" + "-" * (col_width + (i > 0)) if align == "l"
-            else "-" * (col_width + (i > 0)) + ":"
+            (":" + "-" * (col_width - 1 + (i > 0)) + ":") if align == "c"
+            else (":" + "-" * (col_width + (i > 0))) if align == "l"
+            else ("-" * (col_width + (i > 0)) + ":")
             for i, (col_width, align) in enumerate(zip(col_widths, alignments)))]
 
         return "\n".join(lines)
 
-    def csv(self, dialect="excel", csv_kwargs=None):
+    def csv(self,
+            dialect: str = "excel",
+            csv_kwargs: Optional[Dict[str, Any]] = None) -> str:
         """Returns a string containing a CSV representation of the table.
 
         :arg dialect: String passed to :func:`csv.writer`.
@@ -1698,6 +1729,9 @@ class Table:
             10,20
         """
 
+        if not self.rows:
+            return ""
+
         import csv
         import io
 
@@ -1714,17 +1748,80 @@ class Table:
 
         return output.getvalue().rstrip(csv_kwargs["lineterminator"])
 
-    def latex(self, skip_lines=0, hline_after=None):
+    def latex(self,
+            skip_lines: int = 0,
+            hline_after: Optional[Tuple[int, ...]] = None) -> str:
+        r"""Returns a string containing the rows of a LaTeX representation of
+        the table.
+
+        :arg skip_lines: number of lines to skip at the start of the table.
+        :arg hline_after: list of row indices after which to add an ``hline``
+            (the indices must subtract *skip_lines*, if non-zero).
+
+        .. doctest::
+
+            >>> tbl = Table()
+            >>> tbl.add_row([0, "skipped"])
+            >>> tbl.add_row([1, "apple"])
+            >>> tbl.add_row([2, "pear"])
+            >>> print(tbl.latex(skip_lines=1))
+            1 & apple \\
+            2 & pear \\
+        """
+        if not self.rows:
+            return ""
+
         if hline_after is None:
-            hline_after = []
+            hline_after = ()
+
         lines = []
-        for row_nr, row in list(enumerate(self.rows))[skip_lines:]:
-            lines.append(" & ".join(row)+r" \\")
+        for row_nr, row in enumerate(self.rows[skip_lines:]):
+            lines.append(fr"{' & '.join(row)} \\")
             if row_nr in hline_after:
                 lines.append(r"\hline")
 
         return "\n".join(lines)
 
+
+def merge_tables(*tables: Table,
+        skip_columns: Optional[Tuple[int, ...]] = None) -> Table:
+    """
+    :arg skip_columns: a :class:`tuple` of column indices to skip in all the
+        tables except the first one.
+    """
+
+    if len(tables) == 1:
+        return tables[0]
+
+    if any(tables[0].nrows != tbl.nrows for tbl in tables[1:]):
+        raise ValueError("tables do not have the same number of rows")
+
+    if isinstance(skip_columns, int):
+        skip_columns = (skip_columns,)
+
+    def remove_columns(i, row):
+        if i == 0 or skip_columns is None:
+            return row
+        else:
+            return tuple([
+                entry for i, entry in enumerate(row) if i not in skip_columns
+                ])
+
+    alignments = sum([
+        remove_columns(i, tbl._get_alignments())
+        for i, tbl in enumerate(tables)
+        ], ())
+    result = Table(alignments=alignments)
+
+    for i in range(tables[0].nrows):
+        row = []
+        for j, tbl in enumerate(tables):
+            row.extend(remove_columns(j, tbl.rows[i]))
+
+        result.add_row(tuple(row))
+
+    return result
+
 # }}}
 
 
@@ -1773,7 +1870,7 @@ def string_histogram(  # pylint: disable=too-many-arguments,too-many-locals
 
     max_count = max(bins)
     total_count = sum(bins)
-    return "\n".join("%9g |%9d | %3.0f %% | %s" % (
+    return "\n".join("{:9g} |{:9d} | {:3.0f} % | {}".format(
         bin_start,
         bin_value,
         bin_value/total_count*100,
@@ -1791,19 +1888,20 @@ def word_wrap(text, width, wrap_using="\n"):
     breaks are posix newlines (``\n``).
     """
     space_or_break = [" ", wrap_using]
-    return reduce(lambda line, word: "%s%s%s" %
-            (line,
-                space_or_break[(len(line)-line.rfind("\n")-1
-                    + len(word.split("\n", 1)[0])
-                    >= width)],
-                    word),
-            text.split(" ")
-            )
+    return reduce(lambda line, word: "{}{}{}".format(
+        line,
+        space_or_break[(
+            len(line) - line.rfind("\n") - 1
+            + len(word.split("\n", 1)[0])
+            ) >= width],
+        word),
+        text.split(" ")
+        )
 
 # }}}
 
 
-# {{{ command line interfaces -------------------------------------------------
+# {{{ command line interfaces
 
 def _exec_arg(arg, execenv):
     import os
@@ -1827,7 +1925,7 @@ class CPyUserInterface:
         self.doc = doc
 
     def show_usage(self, progname):
-        print("usage: %s <FILE-OR-STATEMENTS>" % progname)
+        print(f"usage: {progname} <FILE-OR-STATEMENTS>")
         print()
         print("FILE-OR-STATEMENTS may either be Python statements of the form")
         print("'variable1 = value1; variable2 = value2' or the name of a file")
@@ -1837,16 +1935,16 @@ class CPyUserInterface:
         print()
         print("The following variables are recognized:")
         for v in sorted(self.variables):
-            print("  {} = {}".format(v, self.variables[v]))
+            print(f"  {v} = {self.variables[v]}")
             if v in self.doc:
-                print("    %s" % self.doc[v])
+                print(f"    {self.doc[v]}")
 
         print()
         print("The following constants are supplied:")
         for c in sorted(self.constants):
-            print("  {} = {}".format(c, self.constants[c]))
+            print(f"  {c} = {self.constants[c]}")
             if c in self.doc:
-                print("    %s" % self.doc[c])
+                print(f"    {self.doc[c]}")
 
     def gather(self, argv=None):
         if argv is None:
@@ -1873,9 +1971,8 @@ class CPyUserInterface:
                 - set(self.constants.keys())):
             if not (added_key.startswith("user_") or added_key.startswith("_")):
                 raise ValueError(
-                        "invalid setup key: '%s' "
-                        "(user variables must start with 'user_' or '_')"
-                        % added_key)
+                        f"invalid setup key: '{added_key}' "
+                        "(user variables must start with 'user_' or '_')")
 
         result = self.Parameters({key: execenv[key] for key in self.variables})
         self.validate(result)
@@ -1919,7 +2016,7 @@ def typedump(val, max_seq=5, special_handlers=None):
         if isinstance(val, dict):
             return "{%s}" % (
                     ", ".join(
-                        "{!r}: {}".format(str(k), typedump(v))
+                        f"{str(k)!r}: {typedump(v)}"
                         for k, v in val.items()))
 
         try:
@@ -1956,8 +2053,7 @@ def invoke_editor(s, filename="edit.txt", descr="the file"):
     else:
         print("(Set the EDITOR environment variable to be "
                 "dropped directly into an editor next time.)")
-        input("Edit %s at %s now, then hit [Enter]:"
-                % (descr, full_name))
+        input(f"Edit {descr} at {full_name} now, then hit [Enter]:")
 
     inf = open(full_name)
     result = inf.read()
@@ -2013,8 +2109,9 @@ class ProgressBar:  # pylint: disable=too-many-instance-attributes
                 self.speed_meas_start_done = self.done
 
             if self.time_per_step is not None:
-                eta_str = "%7.1fs " % max(
-                        0, (self.total-self.done) * self.time_per_step)
+                eta_str = "{:7.1f}s ".format(
+                        max(0, (self.total-self.done) * self.time_per_step)
+                        )
             else:
                 eta_str = "?"
 
@@ -2051,7 +2148,7 @@ class ProgressBar:  # pylint: disable=too-many-instance-attributes
 def assert_not_a_file(name):
     import os
     if os.access(name, os.F_OK):
-        raise OSError("file `%s' already exists" % name)
+        raise OSError(f"file `{name}' already exists")
 
 
 def add_python_path_relative_to_script(rel_path):
@@ -2080,17 +2177,8 @@ def common_dtype(dtypes, default=None):
 
 
 def to_uncomplex_dtype(dtype):
-    import numpy
-    if dtype == numpy.complex64:
-        return numpy.float32
-    elif dtype == numpy.complex128:
-        return numpy.float64
-    if dtype == numpy.float32:
-        return numpy.float32
-    elif dtype == numpy.float64:
-        return numpy.float64
-    else:
-        raise TypeError("unrecgonized dtype '%s'" % dtype)
+    import numpy as np
+    return np.array(1, dtype=dtype).real.dtype.type
 
 
 def match_precision(dtype, dtype_to_match):
@@ -2121,7 +2209,7 @@ def generate_unique_names(prefix):
 
     try_num = 0
     while True:
-        yield "%s_%d" % (prefix, try_num)
+        yield f"{prefix}_{try_num}"
         try_num += 1
 
 
@@ -2135,7 +2223,7 @@ def generate_numbered_unique_names(
         num = 0
 
     while True:
-        name = "%s_%d" % (prefix, num)
+        name = f"{prefix}_{num}"
         num += 1
         yield (num, name)
 
@@ -2176,9 +2264,12 @@ class UniqueNameGenerator:
 
     def add_name(self, name: str) -> None:
         if self.is_name_conflicting(name):
-            raise ValueError("name '%s' conflicts with existing names")
+            raise ValueError(f"name '{name}' conflicts with existing names")
+
         if not name.startswith(self.forced_prefix):
-            raise ValueError("name '%s' does not start with required prefix")
+            raise ValueError(
+                    f"name '{name}' does not start with required prefix "
+                    f"'{self.forced_prefix}'")
 
         self.existing_names.add(name)
         self._name_added(name)
@@ -2203,7 +2294,7 @@ class UniqueNameGenerator:
 
         # }}}
 
-        for counter, var_name in generate_numbered_unique_names(based_on, counter):
+        for counter, var_name in generate_numbered_unique_names(based_on, counter):  # noqa: B007,E501
             if not self.is_name_conflicting(var_name):
                 break
 
@@ -2498,7 +2589,7 @@ class ProcessLogger:  # pylint: disable=too-many-instance-attributes
         fmt_args = [self.description, str(self.timer)]
 
         if extra_msg:
-            msg += ": " + extra_msg
+            msg = f"{msg}: {extra_msg}"
             fmt_args.extend(extra_fmt_args)
 
         self.logger.log(completion_level, msg, *fmt_args)
@@ -2517,17 +2608,22 @@ class DebugProcessLogger(ProcessLogger):
 class log_process:  # noqa: N801
     """A decorator that uses :class:`ProcessLogger` to log data about calls
     to the wrapped function.
+
+    .. automethod:: __init__
+    .. automethod:: __call__
     """
 
-    def __init__(self, logger, description=None):
+    def __init__(self, logger, description=None, long_threshold_seconds=None):
         self.logger = logger
         self.description = description
+        self.long_threshold_seconds = long_threshold_seconds
 
     def __call__(self, wrapped):
         def wrapper(*args, **kwargs):
             with ProcessLogger(
                     self.logger,
-                    self.description or wrapped.__name__):
+                    self.description or wrapped.__name__,
+                    long_threshold_seconds=self.long_threshold_seconds):
                 return wrapped(*args, **kwargs)
 
         from functools import update_wrapper
@@ -2692,6 +2788,95 @@ def unordered_hash(hash_instance, iterable, hash_constructor=None):
 # }}}
 
 
+# {{{ sphere_sample
+
+def sphere_sample_equidistant(npoints_approx: int, r: float = 1.0):
+    """Generate points regularly distributed on a sphere
+    based on https://www.cmu.edu/biolphys/deserno/pdf/sphere_equi.pdf.
+
+    :returns: an :class:`~numpy.ndarray` of shape ``(3, npoints)``, where
+        ``npoints`` does not generally equal *npoints_approx*.
+    """
+
+    import numpy as np
+    points: List[np.ndarray] = []
+
+    count = 0
+    a = 4 * np.pi / npoints_approx
+    d = a ** 0.5
+    M_theta = int(np.ceil(np.pi / d))  # noqa: N806
+    d_theta = np.pi / M_theta
+    d_phi = a / d_theta
+
+    for m in range(M_theta):
+        theta = np.pi * (m + 0.5) / M_theta
+        M_phi = int(np.ceil(2 * np.pi * np.sin(theta) / d_phi))  # noqa: N806
+        for n in range(M_phi):
+            phi = 2 * np.pi * n / M_phi
+            points.append(np.array([
+                r * np.sin(theta) * np.cos(phi),
+                r * np.sin(theta) * np.sin(phi),
+                r * np.cos(theta)
+                ]))
+            count += 1
+
+    # add poles
+    for i in range(3):
+        for sign in [-1, +1]:
+            pole = np.zeros(3)
+            pole[i] = r * sign
+            points.append(pole)
+
+    return np.array(points).T.copy()
+
+
+# NOTE: each tuple contains ``(epsilon, max_npoints)``
+_SPHERE_FIBONACCI_OFFSET = (
+        (0.33, 24), (1.33, 177), (3.33, 890),
+        (10, 11000), (27, 39000), (75, 600000), (214, float("inf")),
+        )
+
+
+def sphere_sample_fibonacci(
+        npoints: int, r: float = 1.0, *,
+        optimize: Optional[str] = None):
+    """Generate points on a sphere based on an offset Fibonacci lattice from [2]_.
+
+    .. [2] http://extremelearning.com.au/how-to-evenly-distribute-points-on-a-sphere-more-effectively-than-the-canonical-fibonacci-lattice/
+
+    :param optimize: takes the values: *None* to use the standard Fibonacci
+        lattice, ``"minimum"`` to minimize the nearest neighbor distances in the
+        lattice and ``"average"`` to minimize the average distances in the
+        lattice.
+
+    :returns: an :class:`~numpy.ndarray` of shape ``(3, npoints)``.
+    """     # noqa: E501
+
+    import numpy as np
+    if optimize is None:
+        epsilon = 0.5
+    elif optimize == "minimum":
+        epsilon, _ = next(o for o in _SPHERE_FIBONACCI_OFFSET if npoints < o[1])
+    elif optimize == "average":
+        epsilon = 0.36
+    else:
+        raise ValueError(f"unknown 'optimize' choice: '{optimize}'")
+
+    golden_ratio = (1 + np.sqrt(5)) / 2
+    n = np.arange(npoints)
+
+    phi = 2.0 * np.pi * n / golden_ratio
+    theta = np.arccos(1.0 - 2.0 * (n + epsilon) / (npoints + 2 * epsilon - 1))
+
+    return np.stack([
+        r * np.sin(theta) * np.cos(phi),
+        r * np.sin(theta) * np.sin(phi),
+        r * np.cos(theta)
+        ])
+
+# }}}
+
+
 def _test():
     import doctest
     doctest.testmod()
